#!/opt/LifeKeeper/bin/perl
#
# Copyright (c) 2004-2009,2011, 2013-2015, 2017-2019, 2021-2022 SIOS Technology Corp.
#
#
# common functions for the NFS Recovery Kit
#
######################## LifeKeeper Linux CHANGE LOG #######################
#
# CR            date    user    description
# ==            ====    ====    ==========
# lklin00003316 01Sep04 PRC     initial submit
# lklin00003359 18Sep04 jef	Fix join for info field
# lklin00003362 18Sep04 jef	Numerous fixes (see CR)
# lklin00003358 20Sep04 jef	Mods for statd running as rpcuser
# lklin00003364 28Sep04 jef	Add exportfs -r when restarting daemons
# lklin00003400 05Nov04 jef	Mods for statd running with -n argument
# lklin00003425 18Nov04 jef	Change exit 1 to return 1
# lklin00003431 19Nov04 jef	Change logic to clear locks even if kernel statd
# lklin00003447 30Dec04 jef	Set hidden info owner during restore&quickCheck
# lklin00003531 04Mar05 nhy     Change lookup_virtual_ips() for better
#                               performance in large, complex hierarchies
# lklin00003524 04Mar05 jef	Only remove hanfs if still in hierarchy
# lklin00003565 24Mar05 jef	Load nfsd for 2.6 and later kernels to force
#				mount of /proc/fs/nfsd
# lklin00003590 27Apr05 jef	Add logic for config option RESTARTMOUNTD
# lklin00003596 03May05 jef	Add lookup logic for number of nfsds to start
# lklin00003614 27May05 jef	Replace exportfs -r with unexport
# lklin00003813 13Feb06 jef	Fixed patern match problems with rmtab info
# lklin00003840 28Mar06 cew	sigHandler needs strings for RKBase preloaded
# lklin00004302 08May07 PRC     fix $? exit codes
# lklin00005016 21Nov08 jef	Cause failover if portmap not running
# bug136        27Mar09 jef	use LKBase::Lookup to get process names
# bug1640       08Mar11 mat	sync data blocks using rename system call
# bug1660       23Mar11 ysa     NFSv4 support
# bug1856       12Jul11 jef     NFSv4 lease and grace not getting set
# bug1831       07Jul11 mat     backup/restore /var/lib/nfs for NFSv4
# bug1840       08Jul11 mat     keep /etc/exports when resource delete
# bug1871       13Jul11 mat     ignore comment of /etc/exports
# bug1909       21Jul11 jef     Only output failed if it really failed
# bug1915       22Jul11 jef     Fix detection of unmounted bind directories
# bug1908       21Jul11 mat     handle multiple fsid=0
# bug1917       22Jul11 jef     Not mounted error should not fail remove
# bug1916       22Jul11 mat     handle multiple bind mount
# bug1919       25Jul11 jef     Fix recovery issue with missing bind mount
# bug1956       29Jul11 jef     Make sure rpc_pipefs is mounted
# bug3434       01Mar13 jef     Add additional checks to ensure NFS is healthy
# bug3459       27Jun13 mat     Fix pingnfs issue with name resolve failure
# bug3278       28Jun14 ysa     Fix hidden directory make failed
# bug7058       17Nov14 ysa     systemd support
# bug7085       18Mar15 ryo     flush and sync before close
# bug7121       09Jul15 ryo     fixed the spelling errors
# bug7122       10Jul15 ryo     Change initial value for same behavior as default without config
# bug7320	21Feb17 kshi	change systemctl argument to nfs-server
# bug7362	28Aug17 ysa     nfs-utils-1.3.0-0.48 support
# bug7493	22Oct18 jef     Fail restore if unable to update exportInfo
# bug7627       21Feb19 ryo     Empty the ROOT env at systemctl enable/disable for SLES15
# bug7747	07Aug19 hoka	Add processing when already unexported in remove
# bug7751	21Aug19 hoka	Change to chach only nfsd daemon with pgrep in stop_daemons
# bug7748	22Aug19 hoka	Add -s option to killall INT
# bug7765       03Sep19 ryo     When protecting NFSv4, add option "-p tcp -v 4" to pingnfs.
# PL-1215       05Dec19 rsa     Add fsid check
# PL-8152       26Feb21	h-ino   Make export() to be recevied from the third arg as export_infos for lkcli import.
# PL-4730       14Sep21 ryo     Unsupport fsid=0
# PL-12327      20Apr22 jef     Add alarm around running exportfs command
########################## end CHANGE LOG ####################################


BEGIN { require '/etc/default/LifeKeeper.pl'; }

package hanfs;
use strict;
use LK;
use LKBase;
use Fcntl qw(:DEFAULT :flock);
use Cwd;
use IO::Handle;

#
# globals
#
my $app;
*app = \('gen');
my $res;
*res= \('hanfs');
my $nfsres;
*nfsres= \('nfs');
my $bindir;
*bindir = \("$ENV{'LKROOT'}/lkadm/subsys/$hanfs::app/$hanfs::res/bin");
my $me;
*me = \(LK::lcduname());

# NFS pathnames
my $etab;
*etab = \(LKBase::Lookup('/var/lib/nfs/etab'));
my $rmtab;
*rmtab = \(LKBase::Lookup('/var/lib/nfs/rmtab'));
my $exports;
*exports = \(LKBase::Lookup('/etc/exports'));
my $lk_exports;
*lk_exports = \('/etc/exports.lk');
my $tmp_exports;
*tmp_exports = \("/tmp/exports$$");
my $export_info_dir;
*export_info_dir = \("$ENV{'LKROOT'}/subsys/$hanfs::app/resources/$hanfs::res/exportInfo/");
my $bind_info_dir;
*bind_info_dir = \("$ENV{'LKROOT'}/subsys/$hanfs::app/resources/$hanfs::res/bindInfo/");
my $mountd_client_list;
*mountd_client_list = \(".mountd_client_list-");
my $statd_client_list;
*statd_client_list = \(".statd_client_list-");
my $smdir;
*smdir = \(LKBase::Lookup('/var/lib/nfs/statd/sm'));
my $smdir_bak;
*smdir_bak = \(LKBase::Lookup('/var/lib/nfs/statd/sm.bak'));
my $alarm_time;
*alarm_time = \(20);
my $callout_script;
*callout_script = \("/etc/default/LifeKeeper-hanfs-callout");
my $failover_locks;
*failover_locks = \((lc("$ENV{'FAILOVERNFSLOCKS'}") eq "true" and ! kernel_resident_statd()) ? 1 : 0);
my $restart_mountd;
*restart_mountd = \(! defined($ENV{'RESTARTMOUNTD'}) || lc("$ENV{'RESTARTMOUNTD'}") eq "true" ? 1 : 0);
my $statedir = "/var/lib/nfs";
my $statedir_bak = "/var/lib/.nfs.LK";

my $debug;
*debug = \(lc("$ENV{'NFS_DEBUG'}") eq "true" ? 1 : 0);
my $mtab = '/etc/mtab';

# list of files that should be present
my @file_list = (
	LKBase::Lookup('/etc/exports'),
);

# list of commands required by hanfs
my @command_list = (
	LKBase::Lookup('/usr/sbin/exportfs'),
);

# list of daemon information used for starting, stopping, and checking NFS
#
# format: [ "process name", "pathname", "options", "required flag", "v4require", "can stop" ]
#
my @daemon_list = (
	[ LKBase::Lookup('portmap'), LKBase::Lookup('/sbin/portmap'), '-w', 1 , 0 , 0 ],  # portmap or rpcbind
	[ LKBase::Lookup('lockd'), LKBase::Lookup('/sbin/rpc.lockd'), '', 0 , 0 , 1 ],
	[ LKBase::Lookup('rpc.statd'), LKBase::Lookup('/sbin/rpc.statd'), '', 0 ,0 , 1 ],
	[ LKBase::Lookup('rpc.rquotad'), LKBase::Lookup('/usr/sbin/rpc.rquotad'), '', 0 , 0 , 1 ],
	[ LKBase::Lookup('rpc.idmapd'), LKBase::Lookup('/usr/sbin/rpc.idmapd'), '-p /var/lib/rpc_pipefs', 0, 1 , 0] ,
	[ gssname(0), gssname(1), '', 0, 1 , 1 ],
	[ LKBase::Lookup('nfsd'), LKBase::Lookup('/usr/sbin/rpc.nfsd'), '8', 1, 1, 0 ],
	[ LKBase::Lookup('rpc.mountd'), LKBase::Lookup('/usr/sbin/rpc.mountd'), '', 1, 1 , 1 ],
);

our $PINGNFS="$ENV{LKROOT}/lkadm/bin/pingnfs";
our $PINGNFSTIME=5;
our $NFSD_RESTART_DELAY_CHECK=((("$ENV{'NFSD_RESTART_DELAY_CHECK'}" ne '') && ("$ENV{'NFSD_RESTART_DELAY_CHECK'}" =~ /(\d+)$/) && ("$ENV{'NFSD_RESTART_DELAY_CHECK'}" > 0)) ? "$ENV{'NFSD_RESTART_DELAY_CHECK'}" : 2);

# Timer for exportfs alarm
our $exportfsTimer=((("$ENV{'NFSEXPORTFSTIMER'}" ne '') && ("$ENV{'NFSEXPORTFSTIMER'}" =~ /(\d+)$/) && ("$ENV{'NFSEXPORTFSTIMER'}" > 0)) ? "$ENV{'NFSEXPORTFSTIMER'}" : 30);

# Action being performed.
our $LKAction;

# For PID cleanup on hangs.
my @psData;
my @PIDKillList;

# Exportfs hung indicator
my $hungExportfs=0;

#
# Common code between hanfs.pm and lk_hanfs_callout
#
require "$ENV{LKROOT}/lkadm/subsys/gen/hanfs/bin/common.pl";

#
# logging functions
#

my $_log = {STDERR => 1}; # log to stderr by default

# pass in a hash reference with the new logging destination(s)
sub logging { # params: \%hash
	my $hash = shift;
	$_log = $hash;
}

sub msg {
	my $pkg = shift;
	my $msg = shift;
	LKBase::LogMessage($_log, $0, $pkg, $msg, @_);
}

sub error {
	my $pkg = shift;
	my $msg = shift;
	LKBase::LogError($_log, $0, $pkg, $msg, @_);
}

sub fatal {
	error @_;
	my ($file, $line) = (caller())[1,2];
	die "Died at: $file:$line\n";
}

sub dbg {
	if ($hanfs::debug) {
		my $msg = shift;
		LKBase::LogMessage($_log, $0, 'RKBase', 'PlainMessage', $msg);
	}
}

#
# Check for kernel resident statd.  On some distributions statd is kernel
# resident and thus we can't failover locks.
#
# returns 1 if kernel resident, 0 otherwise
#
sub kernel_resident_statd {
	my $statd = LKBase::Lookup('rpcKernelStatd');
	my $result;

	($statd eq 'rpcKernelStatd') and return 0;

	chomp ($result = `$statd 2> /dev/null`);

	($result eq 'kernel') and return 1;

	return 0;
}

#
# recovery kit actions (all return 0 for success; non-0 for errors)
#

# perform a health check on the NFS export
sub check { # params: id
	my $id = shift;
	my $export_point = export_point($id);
	my $statd_file = generate_file_name($hanfs::statd_client_list, $id);
	my $hidden_dir = generate_hidden_dir_name($export_point);
	my $ver=3;

	$_ = get_v4info($export_point);
	if (defined($$_{'root'})) {
		$ver = 4;
	}

	if ($hanfs::failover_locks) {
		set_owner_group("$export_point/$hidden_dir", "$export_point/$hidden_dir/$statd_file");
	}

	return (is_nfs_running($id) &&
		is_nfsd_responding(1, $ver) &&
		is_entry_exported($export_point) &&
		is_pipefs_mounted($ver) &&
		is_bind_mounted($export_point) ? 0 : 1);
}

sub stop_daemons { #param: export
	my %stop_list;
	my $export = shift;

	# Set the NFSv4 grace and lease time so they are available on the
	# next start up.  Do this on the stop as it takes a restart of the
	# NFS daemons to get this done.
	set_nfsv4_grace_lease_time();
	foreach my $item (@daemon_list) {
		my ($name) = (@$item)[0];
		$stop_list{$name} = 1 if ((@$item)[5]);
	}
	dbg("stop_daemon: killall -s INT nfsd");
	system('killall -s INT nfsd');
	foreach my $proc (`ps -e`) {
		$proc =~ /^\s*(\d+)\s*(\S+\s*){2}(\S+)$/;
		my $pid = $1;
		my $prog = $3;
		if (defined($stop_list{$prog})) {
			dbg("stop_daemon: kill $prog ($pid)");
			kill(15, $pid);
			waitpid($pid, 0);
		}
	}
	for(my $i = 0; $i < 30; $i++) {
		dbg("stop_daemon: killall -KILL nfsd");
		system('killall -KILL nfsd >/dev/null 2>&1; pgrep -x nfsd >/dev/null');
		last if ($? == (1 << 8));
		sleep 1;
	}
	if ($? != (1 << 8)) {
		error($hanfs::res, 'NFSDstalled', $hanfs::me);
		return 1;
	} else {
		return 0;
	}
}

# unexport the NFS volume
sub remove { # params: id
	my $id = shift;
	my $export_point = export_point($id);

	$LKAction="remove";

	my $rc = unexport($export_point);
	if ($rc != 0) {
		error($hanfs::res, 'UnexportEntryFailed', $export_point, $hanfs::me);
		return $rc;
	}
	my $nfs4 = get_v4info($export_point);
	if (defined($$nfs4{'root'})) {
		# nfsd hold to virtual root.
		# force shutdown
		$rc = stop_daemons();
		return $rc if ($rc != 0);
	}
	$rc = remove_bind($export_point);
	return $rc if ($rc != 0);
	release_locks();
	return $rc;
}

# start NFS (if necessary) and export the NFS volume
sub restore { # params: id
	my $id = shift;
	my $export_point = export_point($id);
	my $v4info = get_v4info($export_point);
	my $ver;
	my $rc;

	$LKAction="restore";

	if (defined($$v4info{'root'})) {
		# $ver = 4;
		my $ins = lk_lookup($id);
		error($hanfs::res, 'UnsupportedFsid0', $ins->{'tag'}, $export_point);
		return 1; # PL-4730: Unsupport fsid=0
	} else {
		$ver = 3;
	}
	dbg("NFS version $ver");

	$rc = restore_bind($export_point);
	dbg("restore_bind($export_point) = $rc");
	if ($rc != 0) {
		return $rc;
	}

	# Check for new entries for the export point before exporting the info
	# from the exportInfo file. No entries would normally be expected.
	my $etabBefore = get_etab_entry($export_point, $ver);

	$rc = restore_nfs($id, 0, $ver); # don't expect it to be exported
	dbg("restore_nfs($id, 0, $ver) = $rc");
	if ($rc != 0) {
		return $rc;
	}

	my $etab = get_etab_entry($export_point, $ver);

	# Only update the exportInfo file if additional export info was found
	# during the above check.
	if (@$etabBefore != 0) {
		if (create_export_info($export_point, $etab, fsid($id))) {
			error($hanfs::res, 'CreateEntryFileFailed', "restore", $export_point);
			return 1;
		}
	}

	if ($hanfs::restart_mountd && !defined($$v4info{'root'})) {
		system("$hanfs::bindir/restart_mountd > /dev/null 2>&1");
		if ($? != 0) {
			error($hanfs::res, 'MountdRestartFailed', $hanfs::me);
			return 1;
		}
	}
}

# local recovery
sub recover { # params: id
	my $id = shift;
	my $export_point = export_point($id);
	my $v4info = get_v4info($export_point);
	my $ver;
	my $rc;

	$LKAction="recover";

	if (defined($$v4info{'root'})) {
		$ver = 4;
	} else {
		$ver = 3;
	}

	$rc = restore_bind($export_point);
	return $rc if ($rc != 0);
	return restore_nfs($id, 1, $ver); # export point should be exported
}

#
# Controlling method for hanfs resource deletion
#
# Returns 0 if no errors; non-0 otherwise
#
sub delete_hanfs { # params: id
	my $id = shift;
	my $ins = lk_lookup($id);
	my $export_point = export_point($id);
	my $shareddir;
	msg('RKBase', 'BeginAction', 'delete', $ins->{'tag'}, $hanfs::me);

	if (-l $statedir) {
		$shareddir = Cwd::realpath("$statedir");
		unlink $statedir;
		dbg("mv -f $statedir_bak $statedir");
		system("mv -f $statedir_bak $statedir");
		if ($? != 0) {
			error($hanfs::res, 'RestoreStatedirFailed', $statedir_bak, $statedir, $hanfs::me);
		}
	}
	if ($ins->{'init'} eq 'AUTORES_ISP') {
		if (restore_exports_entry($export_point) != 0) {
			error($hanfs::res, 'RestoreExportsEntryFailed', $export_point, $hanfs::me);
		}
		if (-d "$shareddir") {
			dbg("cp -a $shareddir/* $statedir/ && sync");
			system("cp -a $shareddir/* $statedir/ && sync");
			system("rm -fr $shareddir > /dev/null 2>&1");
		}
	}

	remove_export_info($export_point);
	remove_bind_info($export_point);

	hanfs::msg('RKBase', 'EndActionSuccess', 'delete', $ins->{'tag'}, $hanfs::me);
	return 0;
}

#
# LK functions
#

# determine whether the NFS export has a corresponding LK resource instance
sub exists_in_lk { # params: id
	my $id = shift;
	my $ins = lk_lookup($id);

	return (defined($ins) && $ins->{'app'} eq $app && $ins->{'res'} eq $res);
}

# instance cache:
my %instance;
# look up (by resource ID) a resource in the LCD database and
# return a reference to a hash structure containing the information
sub lk_lookup { # params: id
	my $id = shift;
	my %hash;
	my @info;

	if (! defined($instance{$id})) {
		@_ = LK::ins_list('', '', '', '', '', $id);
		@_ == 0 and return undef;

		@_ = split /\001/, $_[0];
		$hash{'sys'} = $_[0];
		$hash{'app'} = $_[1];
		$hash{'res'} = $_[2];
		$hash{'tag'} = $_[3];
		@info = split /\002/, $_[5];
		$hash{'info'} = \@info;
		$hash{'state'} = $_[6];
		$hash{'init'} = $_[8];
		$hash{'switchback'} = $_[13];
		$instance{$id} = \%hash;
	}

	return $instance{$id};
}

# generate a unique tag for a new resource (default is the ID)
sub newtag { # params: id
	my $id = shift;

	chomp(my $tag = `$ENV{'LKROOT'}/lkadm/bin/newtag $hanfs::me $id $id`);
	return $tag;
}

# strip "hanfs-" prefix from the resource ID to get the export point
sub export_point { # params: id
	my $id = shift;
	$id =~ s/^hanfs-//;
	return $id;
}

#
# NFS functions
#

# restore_nfs
#
# Called by the restore and recover routines to restart NFS and re-export
# the NFS export point, if necessary
#
# Usage: restore_nfs <id> <was_exported> <ver>
#
# was_exported - set to non-zero to indicate that the export point
# should already be exported
#
# Returns 0 if no errors; 1 otherwise
#
sub restore_nfs { # params: id was_exported ver
	my $id = shift;
	my $was_exported = shift;
	my $ver = shift;
	my $export_point = export_point($id);
	my $return_code;
	my $statd_file = generate_file_name($hanfs::statd_client_list, $id);
	my $hidden_dir = generate_hidden_dir_name($export_point);

	#
	# Attempt to load nfsd and force the mount of /proc/fs/nfsd.
	# On RHEL 4 the mount command below will fail because the
	# mount itself causes the load of nfsd which forces a mount
	# of /proc/fs/nfsd which completes first causing an error.
	#
	system("modprobe nfsd > /dev/null 2>&1");
	if ( -d "/proc/fs/nfsd" && ! -f "/proc/fs/nfsd/exports") {
		system("mount -t nfsd nfsd /proc/fs/nfsd");
		if ($? != 0) {
			error($hanfs::res, 'MountProcNFSD', $hanfs::me);
			return 1;
		}
	}

	if ($hanfs::failover_locks) {
		set_owner_group("$export_point/$hidden_dir", "$export_point/$hidden_dir/$statd_file");
		$return_code=restore_statd_info($id, $export_point);
		($return_code != 0) and return 1;
	}

	if (!is_nfs_running($id)) {
		msg($hanfs::res, 'NFSNotRunning', $hanfs::me);

		#
		# If we are restoring and the export exists in etab we need
		# to clean it up because it will not get re-exported and a
		# stale nfs mount could occur on any connected client.
		#
		if ( ! $was_exported and ! (system("grep -qs \"^$export_point\[[:space:]\]\" $hanfs::etab"))) {
			unexport($export_point);
		}

		if (start_nfs($id) != 0) {
			error($hanfs::res, 'NFSStartFailed', $hanfs::me);
			return 1;
		}
	}

	# Check if nfsd is hung and needs to be restarted
	if (restart_hung_nfsd_processes($ver) == 1) {
		return 1;
	}

	my $already_exported = is_entry_exported($export_point);
	if (!$already_exported) {
		if ($was_exported) {
			msg($hanfs::res, 'EntryNotExported', $export_point, $hanfs::me);
		}

		if (export($export_point) != 0) {
			error($hanfs::res, 'ExportEntryFailed', $export_point, $hanfs::me);
			return 1;
		}
	}

	return 0;
}

# get_etab_entry
#
# Search the $etab file to find all of the entries for $export_point
#
# file format:
#	export_point server(options) server(options) ...
#	(server is optional)
#
# Returns a reference to a list of entries or undef if there were errors
#
sub get_etab_entry { # params: export_point version
	my $export_point = shift;
	my $ver = shift;
	my @etab_entries;
	my $root='';
	my @root_etab_entry;

	open(ETAB, "<$hanfs::etab") or return undef;
	my @etab_file = <ETAB>;
	close ETAB;

	foreach my $etab (@etab_file) {
		# for NFS v3/v4
		if ($etab =~ /^\s*$export_point(\s+|$)/) {
			push(@etab_entries, $etab);
		}
		next if $ver == 3;

		# for NFS v4
		# virtual root not added this time
		$etab =~ /^s*(\S+)\s+(\S+\(\S+\)\s*)+$/;
		my $entry=$1;
		if ($ver == 4 && $2 =~ /fsid\s*=\s*(\d+)/ && $1 == 0) {
			if ($export_point =~ /^$entry/) {
				$root=$entry;
				push(@root_etab_entry, $etab);
				pop(@etab_entries);
			}
		}
	}

	dbg("get_etab_entry: root=\"$root\"");
	# for NFS v4
	if ($root ne '') {
		# export for same root entries
		foreach my $etab (@etab_file) {
			if ($etab =~ /^$root\/.*/) {
				push(@etab_entries, $etab);
			}
		}
		# virtual root entry add to head
		unshift(@etab_entries, @root_etab_entry);
	}

	return \@etab_entries;
}

# is_entry_in_exports
#
# Determine if $export_point is in /etc/exports
#
# Returns 1 if found; 0 if not found
#
sub is_entry_in_exports { # params: export_point
	my $export_point = shift;
	my $found = 0;

	open(EXPORTS, "<$hanfs::exports") or return 0;
	while (<EXPORTS>) {
		if (/^\s*$export_point(\s+|$)/) {
			$found = 1;
			last;
		}
	}
	close EXPORTS;

	return $found;
}

# remove_exports_entry
#
# The /etc/exports file contains all of the directories that are
# automatically exported when NFS is started.
# Therefore, when an NFS export is protected by LifeKeeper, its
# entry must be removed from /etc/exports.
#
# file format:
#	export_point server(options) server(options) ...
#	(server and options are optional)
#
# Returns 0 if no errors; 1 otherwise
#
sub remove_exports_entry { # params: export_point
	my $export_point = shift;
	my $found = 0;
	my $mode;
	my $uid;
	my $gid;
	my $matching = '^\s*' . "$export_point" . '(\s+|$)'; # for NFS v3

	$_ = get_v4info($export_point);
	# if NFSv4, remove all export point under virtual root
	if (defined($$_{'root'})) {
		$export_point = $$_{'root'};
		$matching = '^\s*' . "$export_point"; # for NFS v4

	}
	open(IN, $hanfs::exports) or return 1;
	sysopen(OUT, $hanfs::lk_exports, O_WRONLY | O_CREAT | O_SYNC) or return 1;

	# create $hanfs::lk_exports with the $export_point entry removed
	while (<IN>) {
		if (/$matching/) {
			$found = 1;
			next;
		}
		print OUT $_;
	}

	close OUT;
	close IN;

	if (!$found) { # entry not found so just remove $hanfs::lk_exports
		unlink $hanfs::lk_exports;
	} else {
		# replace $hanfs::exports with $hanfs::lk_exports
		($mode, $uid, $gid) = (stat $hanfs::exports)[2,4,5];

		rename($hanfs::lk_exports, $hanfs::exports) or return 1;
		# get rid of the file system type bits before setting the mode
		chmod ($mode & 07777), $hanfs::exports;
		chown $uid, $gid, $hanfs::exports;
	}

	return 0;
}

# restore_exports_entry
#
# Replaces the export entry that we removed from /etc/exports
# If an entry already exists, place the entry in a temp file
# and log a message
#
# Returns 0 if no errors; 1 otherwise
#
sub restore_exports_entry { # params: export_point
	my $export_point = shift;
	my $entry;
	my $found = 0;
	my @exports;

	my $export_info = read_export_info($export_point);

	open(EXPORTS, "+<$hanfs::exports") or return 1;
	seek EXPORTS, 0, 0;  # go to the beginning of the file
	@exports = <EXPORTS>;
	close(EXPORTS);
	foreach (@exports) {
		if (/^\s*$export_point(\s+|$)/) {
			$found = 1;
			last;
		}
	}

	my $FH_EXPORTS;
	if (!$found) {
		# just append the entry to /etc/exports
		open($FH_EXPORTS, ">$hanfs::exports.LKTMP$$") or return 1;
		foreach (@exports) {
			print $FH_EXPORTS $_;
		}
	} else {
		# conflicting entry found: write the entry to a temp file
		msg($hanfs::res, 'EntryNotAdded', $export_point, $hanfs::tmp_exports);
		open($FH_EXPORTS, ">$hanfs::tmp_exports") or return 1;
	}

	foreach $entry (@$export_info) {
		print $FH_EXPORTS $entry;
	}

	IO::Handle::flush($FH_EXPORTS);
	IO::Handle::sync($FH_EXPORTS);
	close $FH_EXPORTS;
	rename("$hanfs::exports.LKTMP$$", "$hanfs::exports") if (!$found);

	return 0;
}

# strip the first '/' from $export_point and convert the rest to '.'
sub export_info_file { # params: export_point
	my $export_point = shift;

	$export_point =~ s@/@@;
	$export_point =~ s@/@.@g;

	return "$hanfs::export_info_dir$export_point";
}

# create_export_info
#
# Creates the export_info file given the etab file entry (as an array reference)
#
# Returns 0 if no errors; 1 otherwise
#
sub create_export_info { # params: export_point \@etab_entry fsid
	my $export_point = shift;
	my $etab_entry = shift;
	my $fsid = shift;
	my $export_info = export_info_file($export_point);
	my $entry;
	my $new_export_info = "$export_info.LKTMP$$";
	my $error_string;
	my $old_size = 0;
	my $new_size = 0;

	if (! sysopen(EXPORT_INFO, $new_export_info, O_WRONLY | O_CREAT | O_SYNC)) {
		$error_string = $!;
		error($hanfs::res, 'FailedToOpenNewExportInfoFile', $new_export_info, $error_string);
		return 1;
	}

	foreach $entry (@$etab_entry) {
		# Check fsid
		if ($entry !~ /fsid\s*=\s*$fsid/ ) {
			error($hanfs::res, 'FsidDoesNotMatchResourceInfo', $fsid, $hanfs::me);
			close EXPORT_INFO;
			unlink($new_export_info);
			return 1;
		}
		# Make sure the entry is successfully written to the file
		if (! syswrite(EXPORT_INFO, "$entry")) {
			$error_string = $!;
			error($hanfs::res, 'UnableToAddToExportInfoFile', $entry, $new_export_info, $error_string);
			close EXPORT_INFO;
			unlink($new_export_info);
			return 1;
		}
	}
	if (! close(EXPORT_INFO)) {
		$error_string = $!;
		error($hanfs::res, 'CloseFailedForExportInfoFile', $new_export_info, $error_string);
		unlink($new_export_info);
		return 1;
	}

	# Verify the size of the new file is at least as big as the old.
	# If not one of the writes above failed to write all the data.
	(-e $export_info) and ($old_size) = (stat $export_info)[7];
	(-e $new_export_info) and ($new_size) = (stat $new_export_info)[7];
	if ($old_size > $new_size) {
		error($hanfs::res, 'ExportInfoSizeWrong', $new_export_info, $new_size, $old_size);
		unlink($new_export_info);
		return 1;
	}

	if (! rename($new_export_info, $export_info)) {
		$error_string = $!;
		error($hanfs::res, 'UnableToRenameExportInfoFile', $new_export_info, $export_info, $error_string);
		unlink($new_export_info);
		return 1;
	}

	return 0;
}

sub remove_export_info { # params: export_point
	my $export_point = shift;
	my $export_info = export_info_file($export_point);

	unlink($export_info) or return 1;
	return 0;
}

# read_export_info
#
# Returns array reference containing export info file entry, or undef if errors
#
sub read_export_info { # params: export_point
	my $export_point = shift;
	my $export_info = export_info_file($export_point);
	my @entry;

	open(EXPORT_INFO, "<$export_info") or return undef;
	@entry = <EXPORT_INFO>;
	close EXPORT_INFO;
	return \@entry;
}

# copy_export_info
#
# Copies the export_info file from the template system (remote)
# to the target (local) system
#
# Returns 0 if no errors; 1 otherwise
#
sub copy_export_info { # params: id remote
	my $id = shift;
	my $remote = shift;
	my $export_point = export_point($id);
	my $export_info = export_info_file($export_point);
	my $bind_info = bind_info_file($export_point);

	LK::lcdremexec($remote, "lcdrcp $export_info $hanfs::me:$export_info");
	if (! -r $export_info) {
		error($hanfs::res, 'CopyEntryFailed', $export_info, $remote, $hanfs::me);
		return 1;
	}
	$_ = LK::lcdremexec($remote, "[ -f $bind_info ] && echo 1 && lcdrcp $bind_info $hanfs::me:$bind_info");
	if ($_ ne '' && ! -r $bind_info) {
		error($hanfs::res, 'CopyEntryFailed', $bind_info, $remote, $hanfs::me);
		return 1;
	}

	return 0;
}

# parse_entry
#
# Parses the export information and fills a hash using server name
# as the key and the export options for that server as the value
#
# format of each entry is:
# 	export_point server(options) server(options) ...
#	(server is optional)
#
# Returns a reference to a hash containing servers and options
#
sub parse_entry { # params: \@export_entry
	my $export_entry = shift;
	my $entry;
	my %server_options;

	foreach $entry (@$export_entry) {
		# set $export_list to the $entry with the export point removed
		$entry =~ /^\s*\S+(\s+|$)(.*)/ or next;
		my $export_list = $2;

		# break up $export_list into $server and $options
		# and reset $export_list to the remaining part of the list
		while ($export_list =~ /\s*(\S*)\s*\((.*?)\)(.*)/) {
			my $server = $1;
			my $options = $2;
			$export_list = $3;

			if (!$server) {
				if (!$options) {
					# no server, no options, we are done
					last;
				}
				# only have options, set the key to "-"
				$server = '-';
			}
			$server_options{$server} = $options;
		}
	}

	return \%server_options;
}

# compare two NFS export option lists - return 0 if equivalent
sub compare_entries { # params: export_opts1 export_opts2
	my $list1 = shift;
	my $list2 = shift;

	my %opts1 = map { $_ => 1 } (split /,/, $list1);
	my %opts2 = map { $_ => 1 } (split /,/, $list2);

	return LKBase::CompareHash(\%opts1, \%opts2);
}

# is_entry_exported
#
# Returns 1 if entry is exported to all of the expected servers; 0 otherwise
#
sub is_entry_exported { # params: export_point
	my $export_point = shift;
	my $server;
	my $ver = 3;
	my $nfs4 = get_v4info($export_point);
	$ver = 4 if (defined($$nfs4{'root'}));

	# determine which servers $export_point should be exported to
	my $export_info = read_export_info($export_point) or return 0;
	my $export_options = parse_entry($export_info);

	# now determine which servers $export_point is actually exported to
	my $etab = get_etab_entry($export_point, $ver) or return 0;
	my $etab_options = parse_entry($etab);

	# if the entry is not exported to all of the expected servers, return 0
	foreach $server (keys %$export_options) {
		exists $etab_options->{$server} or return 0;
		if (compare_entries($export_options->{$server}, $etab_options->{$server}) != 0) {
			error($hanfs::res, 'MismatchInfoFileAndEtab', $hanfs::me);
			return 0;
		}
	}

	return 1;
}

# returns 0 if no errors; non-0 otherwise
sub export { # params: export_point (export_infos)
	my $export_point = shift;
	my $export_infos = shift; # if lkcli import, read info from lkexport

	return do_export_or_unexport($export_point, 0, $export_infos); # 0 == export
}

# returns 0 if no errors; non-0 otherwise
sub unexport { # params: export_point
	my $export_point = shift;

	return do_export_or_unexport($export_point, 1); # 1 == unexport
}

sub sigHandler { # noparams
	error($hanfs::res, 'ExportfsHung', $LKAction, $LKAction);
	cleanUpPids($$);
	$hungExportfs=1;
	return 1;
};

# do the real work for the export and unexport routines (second param. is 0/1)
# if third param does not exist, read export info from $export_point(first param).
sub do_export_or_unexport { # params: export_point is_unexport (export_infos)
	my $export_point = shift;
	my $is_unexport = shift;
	my $export_infos = shift || read_export_info($export_point) || return 1;
	my $pid;

	$SIG{ALRM} = \&sigHandler;

	# Make sure the script does not timeout for some reason before it has
	# time to do something.

	alarm $exportfsTimer;

	# Fork off the exportfs process so it does not hang the general NFS
	# processing.

	if ($pid = fork) { # parent process
		$hungExportfs=0;
		wait;
		my $parentReturn=$?;
		($hungExportfs > 0) and return 1;
		return ($parentReturn > 0) ? 1 : 0;
	} elsif (!defined $pid) {
		die "$0: Cannot fork: $!\n";
	} else { # child process
		alarm 0;
		for ('HUP','INT','QUIT','ILL','ABRT','BUS','FPE','SEGV','PIPE','TERM','STOP','ALRM','__DIE__') {
			$SIG{$_} = sub {
				die "$0: Process $$ got signal $_[0], process is exiting\n";
			}
		}

		foreach my $export_info (@$export_infos) {
			dbg("do_export_or_unexport: $export_info");
			$export_info =~ /\s*(\S+)(\s+|$).*/;
			$export_point = $1;
			@_ = ($export_info);
			my $export_options = parse_entry(\@_);
			if ($is_unexport) {
				system("exportfs 2>&1 | egrep -q \"^$export_point\s*\"");
				next if ($?);
			}
			if (!defined %$export_options) {
				my $option = $is_unexport ? '-u' : '';
				msg($hanfs::res, 'SystemCall', "exportfs $option :$export_point");
				system "exportfs $option :$export_point";
				exit ($? >> 8);
			}

			while (my ($client, $options) = each %$export_options) {
				if ($is_unexport) {
					$options = '-u';
				} else {
					$options &&= "-i -o $options";
				}
				$client eq '-' and $client = '';

				msg($hanfs::res, 'SystemCall', "exportfs $options $client:$export_point");
				system "exportfs $options $client:$export_point";
				$? != 0 and exit 1;
			}
		}
		exit 0;
	}
}

####################################################################
# getPids(pid)
#
# Description:  Get all children for a given parent process id
#
# Arguments:
#       pid - Parent process id used to find children
#
# Return value:
#       None
#
####################################################################
sub getPids {
	my $pid=shift;
	my $line;

	foreach $line (@psData) {
		# sample (ps line): root   19922 19921 restore
		#   where $_[0] is the process owner
		#         $_[1] is the process id
		#         $_[2] is the parent process id
		chomp ($line);
		@_ = split /\s+/, $line;
		next if ($_[2] != $pid);
		push @PIDKillList, $_[1]; # descendant process
		getPids($_[1]);
	}
}

####################################################################
# cleanUpPids(pid)
#
# Description:  Clean up any child processes of the passed in
#               "pid" that are still running.  These child process
#               may be hung.
#
# Arguments:
#       pid - Process ID to use as the starting point for finding
#               running child processes.
#
# Return value:
#       None
#
####################################################################
sub cleanUpPids {
	my $processID=shift;
	my $ret;

	dbg("Checking for child processes to clean up for pid $processID");
	push @psData, `ps --no-headers -ewwwwf | grep -v \"ps --no-headers\" 2>/dev/null`;
	getPids($processID);

	if (defined @PIDKillList && @PIDKillList != 0 ){
		my $pidlist = join " ", @PIDKillList;
		dbg("These are the pids we will clean up as related to $processID:, $pidlist");

		$ret=kill 15, @PIDKillList;
		if ($ret != @PIDKillList){
			sleep 5;
			error($hanfs::res, 'PidCleanUpKill', $pidlist);
			kill 9, @PIDKillList;
		}
		sleep 5;
	} else {
		dbg("No NFS kit processes found to clean up for pid $processID");
	}
	return 0;
}

# restore_statd_info
#
# Returns 0 if statd info restored, 1 otherwise
#
sub restore_statd_info {
	my $id = shift;
	my $export_point = shift;
	my $statd_info_file;
	my @statd_info=();
	my $file;
	my $lock_file;
	my $error_msg;
	my $hidden_dir_name = generate_hidden_dir_name($export_point);
	my $return_code;

	$statd_info_file = generate_file_name($hanfs::statd_client_list, $id);
	$lock_file = "/tmp/$statd_info_file";
	$statd_info_file = "$export_point/$hidden_dir_name/$statd_info_file";
	if ( -f $statd_info_file ) {
		$return_code = create_lock($lock_file);
		if ($return_code ne "0") {
			fatal($hanfs::res, 'Common', $return_code);
		}
		if (sysopen(STATDINFO, "$statd_info_file", O_RDONLY, 0600) == 0) {
			$error_msg = $!;
			remove_lock($lock_file);
			fatal($hanfs::res, 'OpenFailed', $statd_info_file, $hanfs::me, $error_msg);
		}
		@statd_info = <STATDINFO>;
		close(STATDINFO);
		remove_lock($lock_file);
		foreach (@statd_info) {
			chomp ($file = "$hanfs::smdir_bak/$_");
			sysopen(CLIENT, "$file", O_WRONLY | O_CREAT | O_APPEND | O_SYNC, 0600) or msg($hanfs::res, 'RestoreStatd', $_, $hanfs::me, $!);
			close(CLIENT);
		}
		if (@statd_info != 0 ) {
			system("killall -USR1 rpc.statd");
		}
	}

	return 0;
}

# returns 1 if NFS is installed; 0 otherwise
sub is_nfs_installed { # verno
	my $ver = shift;
	my $item;
	my $ret = 1;

	foreach $item (@file_list) {
		if (! -f $item) {
			error($hanfs::res, 'NFSNotInstalled', $item, $hanfs::me);
			$ret = 0;
		}
	}

	foreach $item (@command_list) {
		if (! -x $item) {
			error($hanfs::res, 'NFSNotInstalled', $item, $hanfs::me);
			$ret = 0;
		}
	}

	foreach $item (@daemon_list) {
		my ($pathname, $v3req, $v4req) = (@$item)[1, 3, 4];

		if (required($ver, $v3req, $v4req) && (! -x $pathname)) {
			error($hanfs::res, 'NFSNotInstalled', $pathname, $hanfs::me);
			$ret = 0;
		}
	}

	return $ret;
}

# returns 1 if NFS daemons are running; 0 otherwise
sub is_nfs_running { # resource id
	my $id = shift;
	my $output;
	my @ip_children;
	my $ip;
	my $ip_hash;
	my $ver;

	my $v4info = get_v4info(export_point($id));
	if (defined($$v4info{'root'})) {
		$ver=4;
	} else {
		$ver=3;
	}

	foreach my $daemon (@daemon_list) {
		my ($process, $v3req, $v4req) = (@$daemon)[0, 3, 4];
		if (required($ver, $v3req, $v4req)) {
			# svcgssd only krb/gss setting.
			next if ($process eq "rpc.svcgssd" and !defined($$v4info{'gss'}));
			`ps -C $process`;
			$? != 0 and return 0;
		} elsif ($process eq "rpc.statd" and $hanfs::failover_locks && !$ENV{'NFS_V4ONLY'}) {
			$output = `ps -wwwfC $process`;
			$? != 0 and return 0;
			$ip_hash = lookup_virtual_ips($id);
			# The following logic will need to be modified when
			# a version of nfs-utils contains a rpc.statd that
			# handles multiple local host names (-n argument).
			foreach $ip (keys %$ip_hash) {
				if ($output =~ /\-n\s*$ip\b/) {
					last;
				} else {
					# rpc.statd is running but not with the
					# hostname arg we need so kill it and
					# return 0 to force a restart of the
					# process via start_nfs
					system("killall -TERM rpc.statd");
					return 0;
				}
			}
		}
	}

	return 1;
}

# returns 1 if NFS daemons are responding; 0 otherwise
sub is_nfsd_responding {
	my $log = shift;
	my $ver = shift; # PL-4730: Ignore this argument
	my $cmd;

	# Check if nfsd is responding
	$cmd="$PINGNFS -s 127.0.0.1 -t $PINGNFSTIME";
	if ($ENV{'NFS_UDP'}) {
		$cmd = "$cmd -p udp";
	} else { # default
		$cmd = "$cmd -p tcp";
	}
	if ($ENV{'NFS_V2V3ONLY'}) {
		$cmd = "$cmd -v 3";
	} else { # default
		$cmd = "$cmd -v 4";
	}

	dbg("$cmd");
	system("$cmd");
	if ($? != 0) {
		error($hanfs::res, 'PingNFSFail', $hanfs::me) if ($log);
		return 0;
	}

	return 1;
}

# Function: set_nfsd_param
#
# Set timeout to nfsv4gracetime and nfsv4leasetime
#
sub set_nfsv4_grace_lease_time {
	my $grace_time = ($ENV{'NFS_V4_GRACE_TIME'} > 0 && $ENV{'NFS_V4_GRACE_TIME'} =~ /(^\d+)$/ ) ? $ENV{'NFS_V4_GRACE_TIME'} : 10;
	my $lease_time = ($ENV{'NFS_V4_LEASE_TIME'} > 0 && $ENV{'NFS_V4_LEASE_TIME'} =~ /(^\d+)$/ ) ? $ENV{'NFS_V4_LEASE_TIME'} : 10;
	if (-f '/proc/fs/nfsd/nfsv4gracetime') {
		sysopen(PROCFS, '/proc/fs/nfsd/nfsv4gracetime', O_WRONLY);
		 $_ = syswrite(PROCFS, $grace_time, 2);
		dbg("nfsv4gracetime set failed $_") if ($_ != 2);
		close(PROCFS);
	}
	if (-f '/proc/fs/nfsd/nfsv4leasetime') {
		sysopen(PROCFS, '/proc/fs/nfsd/nfsv4leasetime', O_WRONLY);
		$_ = syswrite(PROCFS, $lease_time, 2);
		dbg("nfsv4leasetime set failed $_") if ($_ != 2);
		close(PROCFS);
	}
}

# Function: start_nfs
#
# Starts the NFS daemons (non-critical daemons are started if they exist)
#
# Returns 0 if no errors; 1 otherwise
#
sub start_nfs { # resource id
	my $id = shift;
	my $ins = lk_lookup($id);
	my $ip_hash;
	my $ip;
	my $host_arg;
	my $lookup_string;
	my $lookup_results;
	my $ver;
	my $gss;

	$ver = 3;

	$_ = get_v4info(export_point($id));
	if (defined($$_{'root'})) {
		$ver = 4;
	}
	dbg("nfs version $ver");
	$gss = defined($$_{'gss'});
	# Try setting grace and lease time before starting any NFS process.
	# To set them requires a stop and restart of NFS so we may be able
	# to set them on this initial start since NFS did not appear to be
	# running.
	set_nfsv4_grace_lease_time() if (!$ENV{'NFS_V2V3ONLY'});

	foreach my $daemon (@daemon_list) {
		my ($process, $pathname, $options, $v3req, $v4req) = (@$daemon);
		`ps -C $process`;
		dbg("start_nfs: $process $?");
		if ($? != 0) {
			if (required($ver, $v3req, $v4req)) {
				# go ahead.
			} elsif ($process eq "rpc.statd" && !$ENV{'NFS_V4ONLY'}) {
				if ($hanfs::failover_locks) {
					$ip_hash = lookup_virtual_ips($id);
					# The following logic will need to be
					# modified when a version of nfs-utils
					# contains a rpc.statd that handles
					# multiple local host names
					# (-n argument).
					foreach $ip (keys(%$ip_hash)) {
						$host_arg = "-n $ip";
					}
					$options = "$options -H $hanfs::callout_script $host_arg";
				} else {
					next;
				}
			} else {
				next;
			}

			if (! -x $pathname) {
				return 1;
			}
			next if ($process eq "rpc.svcgssd" && !$gss);
			dbg("daemon \"$process\" starting...");

			if ($process eq "nfsd") {
				$lookup_string=LKBase::Lookup("nfsdCount");
				chomp ($lookup_results=`$lookup_string`);
				if ($lookup_results ne '') {
					$options = $lookup_results;
				}
			}
			if ($process eq "rpc.idmapd" || $process eq "rpc.svcgssd") {
				# Make sure rpc_pipefs is mounted
				if (! mount_pipefs($ver)) {
					return 1;
				}
			}
			msg($hanfs::res, 'SystemCall', "$pathname $options");
			system("$pathname $options");
			dbg("exitcode: $?");
			return 1 if ($? != 0);
		}
	}

	return 0;
}

# Function: restart_hung_nfsd_processes
#
# Stops and restarts hung nfsd processes.
#
# Returns 0 if nfsd is successfully restarted, 1 if it failed to restart and
# respond to ping attempts.
#
sub restart_hung_nfsd_processes { # params: ver
	my $ver = shift;
	my $cmd;

	if (!is_nfsd_responding(0, $ver)) {
		msg($hanfs::res, 'KillHungNFSD', $hanfs::me);
		$cmd="killall -q -9 nfsd";
		system("$cmd");
		if (start_nfs() != 0) {
			error($hanfs::res, 'NFSStartFailed', $hanfs::me);
			return 1;
		}
		sleep ($NFSD_RESTART_DELAY_CHECK);
		if (!is_nfsd_responding(0, $ver)) {
			error($hanfs::res, 'NFSDRestartFailed', $hanfs::me);
			return 1;
		}
	}

	return 0;
}

# returns the fsid from the resource info field
sub fsid { # params: id
	my $id = shift;
	my $ins = lk_lookup($id);

	return ${$ins->{'info'}}[1];
}

# creates a unique (to the local system) fsid and returns it
sub gen_fsid {
	my $export_point = shift;
	my $fsid;

	do {
		$fsid = sprintf("%u", int(rand(0x7fffffff)));
	} while (fsid_exists($fsid, $export_point));

	return $fsid;
}

# returns 1 if a resource or /etc/exports entry is already using the fsid;
# 0 if not
sub fsid_exists { # params: fsid
	my $fsid = shift;
	my $export_point = shift;
	my @results=();
	my $string="fsid=$fsid";

	@_ = LK::ins_list('', '', 'gen', 'hanfs', '', '');
	foreach (@_) {
		@_ = split /\001/, $_;
		@_ = split /\002/, $_[5];
		return 1 if ($fsid eq $_[1]);
	}

	# Check /etc/exports
	open (EXPORTS, "</etc/exports") or fatal($hanfs::res, 'OpenFailed', "/etc/exports", $hanfs::me, $!);
	@results = grep (!/^$export_point\s/, <EXPORTS>);
	(grep (/^[^#]+$string\b/, @results)) and return 1;

	return 0;
}

# release_locks
#
# Tells lockd to release all file locks so that we can unmount the filesystem
#
sub release_locks { # noparams
	my $kernelStatd;

	system('killall -9 lockd');
	$kernelStatd = kernel_resident_statd();
	if (! $kernelStatd) {
		system("killall -USR1 rpc.statd");
	}
}

# get_hanfs_id_for_export
#
# Find the hanfs resource instance ID given the export point
#
# Returns instance ID if ISP or undef.
sub get_hanfs_id_for_export {
	my $export_point = shift;
	my $id = "hanfs-$export_point";
	my $ins_info;
	my $state;

	$ins_info=LK::ins_list('', '', '', '', '', $id);

	$state = (split /\x01/, $ins_info)[6];

	($state eq "ISP") and return $id;

	return (undef);
}

# get_all_hanfs
#
# Find all hanfs resource instance export points and instance IDs
#
# Returns a reference to a hash structure containing the export point and
# hanfs instance id (hash value)
sub get_all_hanfs {
	my @ins_info=();
	my $ins;
	my $export_dir;
	my $id;
	my $state;
	my %hash;

	@ins_info=LK::ins_list('', '', 'gen', 'hanfs', '', '');

	foreach $ins (@ins_info) {
		($id, $state) = (split /\x01/, $ins)[4,6];
		$export_dir = (split /\x02/, (split /\x01/, $ins)[5])[0];
		if ($state eq "ISP") {
			$hash{$export_dir} = $id;
		}
	}

	return \%hash;
}

# populate_statd_info
#
# Populate the hidden statd info file with current info from sm directory
# (if any)
#
# Returns 0 if no errors; non-0 otherwise
sub populate_statd_info {
	my $export_point = shift;
	my $statd_info_file = shift;
	my $hidden_dir_name = generate_hidden_dir_name($export_point);
	my $file = "$export_point/$hidden_dir_name/$statd_info_file";
	my $lock_file = "/tmp/$statd_info_file";
	my $error_msg;
	my @dir_entries=();
	my $return_code;

	$return_code = create_lock($lock_file);
	if ($return_code ne "0") {
		error($hanfs::res, 'Common', $return_code);
		return 1;
	}

	if (!opendir (SMDIR, $hanfs::smdir)) {
		remove_lock($lock_file);
		error($hanfs::res, 'OpenFailed', $hanfs::smdir, $hanfs::me, $error_msg);
		return 1;
	}

	@dir_entries = grep !/^\./, readdir SMDIR;
	closedir (SMDIR);

	if ( ! -d "$export_point/$hidden_dir_name" ) {
		if (mkdir("$export_point/$hidden_dir_name", 0700) == 0) {
			$error_msg = $!;
			remove_lock($lock_file);
			error($hanfs::res, 'MakeDirectory', "$export_point/$hidden_dir_name", $hanfs::me, $error_msg);
			return 1;
		}
	}

	if (sysopen(STATDINFO, "$file.LKTMP$$", O_WRONLY | O_CREAT | O_TRUNC | O_SYNC, 0600) == 0) {
		$error_msg = $!;
		remove_lock($lock_file);
		error($hanfs::res, 'OpenFailed', $file, $hanfs::me, $error_msg);
		return 1;
	}
	foreach (@dir_entries) {
		print STATDINFO "$_\n";
	}

	close (STATDINFO);
	rename("$file.LKTMP$$", "$file");
	set_owner_group("$export_point/$hidden_dir_name", $file);
	remove_lock($lock_file);

	return 0;
}

#
# Set owner and group on hidden dir and statd info file
#
sub set_owner_group {
	my $hidden_dir = shift;
	my $hidden_file = shift;
	my $uid;
	my $gid;

	($uid, $gid) = (stat $hanfs::smdir)[4,5];
	chown $uid, $gid, $hidden_dir;
	chown $uid, $gid, $hidden_file;
}


#
# Controlling method for deletion of the nfs resource hierarchy
#
sub delete_nfs {
	my $nfs_tag = shift;
	my $nfs_id = shift;
	my $delete_info = "DELETEME";
	my @ip_deps=();
	my $ip_child;
	my $ip_child_tag;
	my $hanfs_id = 'ha' . $nfs_id;
	my @hanfs_deps=();
	my $hanfs_child;
	my $hanfs_child_tag;
	my $hanfs_child_id;
	my $nfs_ins;
	my $hanfs_ins;
	my @filesys_deps=();
	my $filesys_child;
	my $filesys_child_tag;
	my $ins;
	my $eqv_info;
	my $new_info;
	my @eqvs = ();
	my $eqv_resource;
	my $nfs_eqv_sys;
	my $nfs_eqv_tag;
	my $nfs_eqv_id;
	my $hanfs_eqv_id;
	(my $export_point = $nfs_id) =~ s/^nfs-//;
	my $do_final_remove=0;
	my $mountd_info_file = generate_file_name($hanfs::mountd_client_list, $hanfs_id);
	my $statd_info_file = generate_file_name($hanfs::statd_client_list, $hanfs_id);
	my $hidden_dir_name = generate_hidden_dir_name($export_point);

	msg('RKBase', 'BeginAction', 'delete', $nfs_tag, $hanfs::me);

	$ins = lk_lookup($nfs_id);
	if ($ins->{'info'}[0] eq $delete_info) {
		hanfs::msg('RKBase', 'EndActionSuccess', 'delete', $nfs_tag, $hanfs::me);
		return 0;
	}

	@ip_deps = LK::dep_list($hanfs::me, $nfs_tag, '', 'comm', 'ip');
	foreach $ip_child (@ip_deps) {
		$ip_child_tag = (split /\x01/, $ip_child)[1];
		@hanfs_deps = LK::dep_list($hanfs::me, $ip_child_tag, '', 'gen', 'hanfs');
		foreach $hanfs_child (@hanfs_deps) {
			$hanfs_child_tag = (split /\x01/, $hanfs_child)[1];
			$hanfs_ins = LK::ins_list($hanfs::me, '', '', '', $hanfs_child_tag, '');
			$hanfs_child_id = (split /\x01/, $hanfs_ins)[4];
			if ($hanfs_id eq $hanfs_child_id) {
				$do_final_remove=1;
				LK::dep_remove($hanfs::me, $ip_child_tag, $hanfs_child_tag);
				@filesys_deps = LK::dep_list($hanfs::me, $hanfs_child_tag, '', 'gen', 'filesys');
				foreach $filesys_child (@filesys_deps) {
					$filesys_child_tag = (split /\x01/, $filesys_child)[1];
#					LK::ins_mark($filesys_child_tag, 'r', 0);
					LK::dep_remove($hanfs::me, $hanfs_child_tag, $filesys_child_tag);
				}
#				LK::ins_remove($hanfs::me, $hanfs_child_tag, '', '', '', '');
#				LK::ins_remove($hanfs::me, '', '', '', $hanfs_child_tag, '');
			} else {
				LK::ins_mark($hanfs_child_tag, 'r', 0);
			}
		}
		LK::ins_mark($ip_child_tag, 'r', 0);
	}

	#
	# Mark delete in progress
	#
	$new_info = join "\x02", ($delete_info, @{$ins->{'info'}});
	LK::ins_setinfo($hanfs::me, $nfs_id, $new_info);

	@eqvs = LK::eqv_list($hanfs::me, $hanfs::me, $nfs_tag);
	LK::lcderrclr();
	foreach $eqv_resource (@eqvs) {
		($nfs_eqv_sys, $nfs_eqv_tag) = (split /\x01/, $eqv_resource)[2,3];
		$nfs_ins = LK::ins_list($nfs_eqv_sys, '', '', '', $nfs_eqv_tag, '');
		($nfs_eqv_id, $eqv_info) = (split /\x01/, $nfs_ins)[4,5];
		$new_info = join "\x02", ($delete_info, $eqv_info);
		LK::ins_setinfo($nfs_eqv_sys, $nfs_eqv_tag, $new_info);
		@ip_deps = LK::dep_list($nfs_eqv_sys, $nfs_eqv_tag, '', 'comm', 'ip');
		foreach $ip_child (@ip_deps) {
			$ip_child_tag = (split /\x01/, $ip_child)[1];
			@hanfs_deps = LK::dep_list($nfs_eqv_sys, $ip_child_tag, '', 'gen', 'hanfs');
			LK::lcderrclr();
			foreach $hanfs_child (@hanfs_deps) {
				$hanfs_child_tag = (split /\x01/, $hanfs_child)[1];
				$hanfs_ins = LK::ins_list($nfs_eqv_sys, '', '', '', $hanfs_child_tag, '');
				$hanfs_child_id = (split /\x01/, $hanfs_ins)[4];
				$hanfs_eqv_id = "ha" . $nfs_eqv_id;
				if ($hanfs_eqv_id eq $hanfs_child_id) {
					LK::dep_remove($nfs_eqv_sys, $ip_child_tag, $hanfs_child_tag);
					LK::lcderrclr();
					@filesys_deps = LK::dep_list($nfs_eqv_sys, $hanfs_child_tag, '', 'gen', 'filesys');
					foreach $filesys_child (@filesys_deps) {
						$filesys_child_tag = (split /\x01/, $filesys_child)[1];
						LK::dep_remove($nfs_eqv_sys, $hanfs_child_tag, $filesys_child_tag);
					}
					LK::ins_remove($nfs_eqv_sys, $hanfs_child_id, '', '', '', '', 0, 0, 0, 0);
				}
			}
			LK::dep_remove($nfs_eqv_sys, $nfs_eqv_tag, $ip_child_tag);
		}
		LK::ins_remove($nfs_eqv_sys, $nfs_eqv_tag, '', '', '', '', 0, 0, 0, 0);
		LK::lcdsync($nfs_eqv_sys);
	}

	# Find hanfs and clean its remants up but only if it has dependents
	if ($do_final_remove) {
		$hanfs_ins = LK::ins_list($hanfs::me, '', '', '', '', $hanfs_id);
		$hanfs_child_tag = (split /\x01/, $hanfs_ins)[3];
		LK::ins_remove($hanfs::me, $hanfs_child_tag, '', '', '', '');
		LK::lcderrclr();
	}

	# cleanup hidden files and directory
	(-f "$export_point/$hidden_dir_name/$mountd_info_file") and unlink ("$export_point/$hidden_dir_name/$mountd_info_file");
	(-f "$export_point/$hidden_dir_name/$statd_info_file") and unlink ("$export_point/$hidden_dir_name/$statd_info_file");

	rmdir ("$export_point/$hidden_dir_name");

	LK::lcdsync($hanfs::me);

	hanfs::msg('RKBase', 'EndActionSuccess', 'delete', $nfs_tag, $hanfs::me);
	return 0;
}

#
# Looks up the IP addresses of the hanfs parent IP resources in the hierarchy
# plus any other IP resources that are parents to ISP hanfs resources.
#
# Returns a reference to a hash of IP addresses or undef if an error occurs or
# it finds more than one IP (rpc.statd in the current nfs-utils package does
# not support more than one local host name via the -n argument at this time).
#
sub lookup_virtual_ips {
	my $id = shift;
	my @ip_instances;
	my $ip_ins_id;
	my $ip_tag;
	my $ip_ins;
	my $ip;
	my %ip_hash;
	my @hanfs_children;
	my $hanfs_child;
	my $hanfs_tag;
	my $hanfs_ins;
	my $hanfs_state;
	my $ins = lk_lookup($id);

	# Get all IP resources and look for those that are hanfs parents
	@ip_instances = LK::ins_list($hanfs::me, '', 'comm', 'ip', '', '');
	foreach $ip_ins (@ip_instances) {
		($ip_tag, $ip_ins_id) = (split /\x01/, $ip_ins)[3,4];
		($ip = $ip_ins_id) =~ s/^IP-//;

		# Get list of hanfs children of this IP
		@hanfs_children = LK::dep_list($hanfs::me, $ip_tag, '', 'gen', 'hanfs');

		# Look for current hanfs resource in the list
		foreach $hanfs_child (@hanfs_children) {
			$hanfs_tag = (split /\x01/, $hanfs_child)[1];
			if ($hanfs_tag eq $ins->{'tag'}) {
				$ip_hash{"$ip"} = $ip;
				last;
			}
		}

		# If current hanfs resource wasn't found, look for any
		# hanfs children that are ISP
		if (!defined($ip_hash{"$ip"})) {
			foreach $hanfs_child (@hanfs_children) {
				$hanfs_tag = (split /\x01/, $hanfs_child)[1];
				$hanfs_ins = LK::ins_list($hanfs::me, '', '', '', $hanfs_tag, '');
				$hanfs_state = (split /\x01/, $hanfs_ins)[6];
				if ($hanfs_state eq "ISP") {
					$ip_hash{"$ip"} = $ip;
					last;
				}
			}
		}
	}

	if (keys(%ip_hash) gt 1) {
		# We have a problem because we can't do more than one ip with
		# rpc.statd in the current release of the nfs-utils package.
		# Log a message and continue running rpc.statd without lock
		# failover.  This check can be removed with rpc.statd allows
		# more than one local host name for the -n argument.
		error($hanfs::res, 'NoMultipleHosts');
		return undef;
	}

	return \%ip_hash;
}

sub required {
	my $resver = shift; # PL-4730: Ignore this argument
	my $v3req = shift;
	my $v4req = shift;

	return $v3req if ($ENV{'NFS_V2V3ONLY'});
	return $v4req if ($ENV{'NFS_V4ONLY'});
	return ($v3req || $v4req);
}

sub get_v4info {
	my $export_point = shift;
	my %ret;
	my $export_infos = read_export_info($export_point);
	foreach my $export_info (@$export_infos) {
		$export_info =~ /\s*(\S+)(\s+|$).*/;
		$export_point = $1;
		@_ = ($export_info);
		my $export_options = parse_entry(\@_);
		foreach my $host (keys %$export_options) {
			my $options = $$export_options{$host};
			$ret{'root'} = $export_point if ($options =~ /fsid=0/);
			$ret{'gss'} = 1 if ($host =~ /^gss\/krb5/);
		}
	}
	return \%ret;
}

sub get_virtual_root { #param: etab files
	foreach (@_) {
		/^\s*(\S+)(\s*|$)/ || next;
		my $export = $1;
		/^s*\S+\s+(\S+\(\S+\)\s*)+$/;
		# "fsid=0" is NFSv4 virtual root
		if ($1 =~ /fsid\s*=\s*(\d+)/ && $1 == 0) {
			return $export;
			next;
		}
	}
	return '';
}

sub export_version { # param: export point
	my $export_point=shift;
	open EXPORTS, "<$hanfs::etab" || return 0;
	my @exports=<EXPORTS>;
	close(EXPORTS);

	my $root=get_virtual_root(@exports);
	if  ($root ne '') {
		if ($export_point =~ /^$root/) {
		    	return 4;
		} else {
			return 3;
		}
	}
}

sub getexportfs { #param: export point
	my $export_point=shift;
	my %v4export;
	my @exportfs=($export_point);
	my %mount;
	my $root;

	open EXPORTS, "<$hanfs::etab" || return 0;
	my @exports=<EXPORTS>;
	close(EXPORTS);

	$root=get_virtual_root(@exports);

	if ($root eq '' || $root ne $export_point) {
		# NFSv2/v3 export
		return @exportfs;
	}

	# lookup v4 exported entry
	foreach (@exports) {
		/^\s*(\S+)(\s*|$)/ || undef;
		my $export = $1;
		$v4export{$export} = 1 if ($export =~ /^$root\/.*/);
	}

	# lookup bind mount entry
	open MTAB, "<$mtab" || return undef;
	while (<MTAB>) {
		split / /;
		$mount{$_[1]} = $_[0] if ($_[3] =~ /bind/);
	}
	close(MTAB);


	foreach (keys %v4export) {
		if (defined $mount{$_}) {
			push (@exportfs, $mount{$_});
		}
	}
	return @exportfs;
}

sub bind_info_file { # params: export_point
	my $export_point = shift;

	$export_point =~ s@/@@;
	$export_point =~ s@/@.@g;

	return "$hanfs::bind_info_dir$export_point";
}

sub remove_bind_info { # params: export_point
	my $export_point = shift;
	my $bind_info = bind_info_file($export_point);

	unlink($bind_info) or return 1;
	return 0;
}

sub restore_bind { #param: export_point
	return do_mount_bind($_[0], 0);
}

sub remove_bind { #param: export_point
	return do_mount_bind($_[0], 1);
}

sub dir_depth { #param: directory
	my $dir = shift;
	my @d;
	while ($dir =~ /(\/\S*?)/g) {
	    push @d, $1;
	}
	return @d;
}

#sort for directory depth
sub sort_depth { #param: mount entry sortorder
	my $entry = shift;
	my $order = shift;
	@$entry = sort {
		if ($order) {
			dir_depth($b) - dir_depth($a);
		} else {
			dir_depth($a) - dir_depth($b);
		}
	} @$entry;
}

sub do_mount_bind { #param: export_point umount
	my $export_point = shift;
	my $umount = shift;
	my $bindinfo = bind_info_file($export_point);
	my %mtab;
	my $ret = 0;

	open MTAB, "<$mtab" || return 1;
	while (<MTAB>) {
	    split;
	    $mtab{$_[1]} = $_[0];
	}
	close(MTAB);

	open BIND, "<$bindinfo" || return 1;
	my @bindmount = <BIND>;
	close(BIND);
	sort_depth(\@bindmount, $umount);
	foreach (@bindmount) {
	    	chomp;
	    	(my $dev, my $point ) = split /\x01/;
		if ($umount == 0) {
			if ( ! -d $point ) {
				system("mkdir -p $point");
			}
			if ($mtab{$point} eq $dev) {
				# already mounted.
				next;
			} elsif (defined($mtab{$point})) {
				$ret = system("umount $point");
				return $ret >> 8 if ($ret != 0);
			}
			dbg("mount -o bind $dev $point");
			$ret = system("mount -o bind $dev $point");
		} else {
			dbg("umount $point");
			if (defined $mtab{$point}) {
				$ret = system("umount $point");
				if ($ret != 0) {
					error($hanfs::res, 'BindUmountFailed', $point, $export_point, $hanfs::me);
				}
			}
		}
		return $ret >> 8 if ($ret != 0);
	}
	return 0;
}

sub is_bind_mounted { #param: export_point
	my $bindinfo = bind_info_file($_[0]);
	my %mountinfo;
	my $ret = 1;
	my $dev;
	my $point;

	open MTAB, "<$mtab" || return 0;
	while (<MTAB>) {
		chomp;
		($dev, $point) = split;
		$mountinfo{$point} = $dev;
	}
	close(MTAB);

	open BIND, "<$bindinfo" || return 0;
	while (<BIND>) {
	    	chomp;
	    	($dev, $point ) = split /\x01/;
		if ((!defined $mountinfo{$point}) || ($mountinfo{$point} ne $dev)) {
			$ret=0;
			last;
		}
	}
	close(BIND);
	return $ret;
}

# mount_pipefs
#
# Mount rpc_pipefs if it is not already mounted
#
# Returns 1 if mounted successfully or 0 if mount failed
sub mount_pipefs { #param: $ver
	my $ver = shift;
	my $cmd = "mount -t rpc_pipefs sunrpc /var/lib/rpc_pipefs 2>&1";
	my @results;

	return 1 if (is_pipefs_mounted($ver));

	chomp (@results=`$cmd`);
	if ($? != 0) {
		msg($hanfs::res, 'PipeFSMountFail', $hanfs::me, @results);
		return 0;
	}

	msg($hanfs::res, 'MountedPipeFS', $hanfs::me);
	return 1;
}

# is_pipefs_mounted
#
# Check if rpc_pipefs is currently mounted
#
# Returns 1 if rpc_pipefs is currently mounted; 0 if not currently mounted
sub is_pipefs_mounted { #param: $ver
	my $ver = shift; # PL-4730: Ignore this argument
	# How check is done in rc script on RedHat
	my $cmd="grep -v \"^#\" /proc/mounts | awk '{ if (\$3 ~ \/^rpc_pipefs\$\/ ) print \$2}'";
	my @results;

	return 1 if ($ENV{'NFS_V2V3ONLY'}); # Only need to check if v4
	chomp (@results=`$cmd 2>/dev/null`);
	dbg("rpc_pipefs is mounted results: @results");
	foreach (@results) {
		return 1 if ($_ eq "/var/lib/rpc_pipefs");
	}
	dbg("rpc_pipefs is not mounted");
	return 0;
}

sub gssname {
	my $type = shift;
	my $bin = LKBase::Lookup("/usr/sbin/gssproxy");
	my @old_nfs = ( LKBase::Lookup("rpc.svcgssd"), LKBase::Lookup("/usr/sbin/rpc.svcgssd") );
	my @new_nfs = ( LKBase::Lookup("gssproxy"), LKBase::Lookup("/usr/sbin/gssproxy") );
	if ( -x $bin) {
		return $new_nfs[$type];
	} else {
		return $old_nfs[$type];
	}
}

1;
